#include "grf.h"
#if NGRF > 0

/* Graphics routines for the AMIGA native custom chip set. */

/* NOTE: this is now only a more or less rough interface to Chris Hopps'
         view driver. Due to some design problems with RTG view can't
	 currently replace grf, but that's an option for the future. */

#include "param.h"
#include "vm/vm_param.h"
#include "sys/errno.h"
#include "ioctl.h"
#include "grfioctl.h"
#include "grfvar.h"
#include "grf_ccreg.h"
#include "../include/cpu.h"
#include "../amiga/custom.h"
#include "../amiga/cia.h"
#include "../amiga/cc_types.h"
#include "viewioctl.h"

/* Initialize hardware.
 * Must point g_display at a grfinfo structure describing the hardware.
 * Returns 0 if hardware not present, non-zero ow.
 */
cc_init(gp, ad)
	struct grf_softc *gp;
	struct amiga_device *ad;
{
  struct grfinfo *gi = &gp->g_display;
  int rc;
  long start, off;
  struct view_size vs;

  /* already initialized fail */
  if (gp->g_data == 12345678)
    return 0;

  rc = grfcc_probe ();
  if (! rc)
    return 0;

  viewprobe ();

  /* now it gets REALLY nasty.. view[0] stays open for all time
     now.. */
  if (viewopen (0, 0))
    return 0;

  /* this torture is needed to query start and length of
     mappable bitmap. CH please do something about this... */
  off   = 0;
  start = viewmap (0, off, 0); /* protections are ignored for this purpose */
  if (start == -1)
    return 0;
  for (off += NBPG; viewmap (0, off, 0) != -1; off += NBPG) ;

  gp->g_data = 12345678;	/* not particularly clean.. */
  gi->gd_regaddr = 0xdff000;
  gi->gd_regsize = round_page(sizeof (custom));

  gi->gd_fbaddr  = start;
  gi->gd_fbsize  = off - start;

  if (viewioctl (0, VIEW_GETSIZE, &vs, 0, 0))
    {
      /* fill in some default values... XXX */
      vs.width = 640; vs.height = 400; vs.depth = 2;
    }

  gi->gd_colors = 1 << vs.depth;
  gi->gd_planes = vs.depth;
  
  /* sigh, no more oversized framebuffers for cc-console :-(( */
  gi->gd_fbwidth  = vs.width;
  gi->gd_fbheight = vs.height;
  gi->gd_fbx	  = 0;
  gi->gd_fby	  = 0;
  gi->gd_dwidth   = vs.width;
  gi->gd_dheight  = vs.height;
  gi->gd_dx	  = 0;
  gi->gd_dy	  = 0;
  
  gp->g_regkva = 0;	/* builtin */
  gp->g_fbkva  = 0;	/* not needed, view internal */

  viewioctl (0, VIEW_DISPLAY, 0, 0, 0);

  return(1);
}

cc_config(gp, di)
	register struct grf_softc *gp;
	struct grfdyninfo *di;
{
  struct grfinfo *gi = &gp->g_display;

  /* bottom missing... */

}

/*
 * Change the mode of the display.
 * Right now all we can do is grfon/grfoff.
 * Return a UNIX error number or 0 for success.
 */
cc_mode(gp, cmd, arg)
	register struct grf_softc *gp;
	int cmd;
	void *arg;
{
  switch (cmd)
    {
    case GM_GRFON:
      return 0;
      
    case GM_GRFOFF:
      return 0;
      
    case GM_GRFCONFIG:
      return cc_config (gp, (struct grfdyninfo *) arg);

    default:
      break;
    }
    
  return EINVAL;
}

#endif

