#include "ite.h"
#if ! defined (NITE)
#define NITE 1
#endif
#if NITE > 0

#include "param.h"
#include "conf.h"
#include "proc.h"
#include "ioctl.h"
#include "tty.h"
#include "systm.h"

#include "ite.h"
#include "itevar.h"
#include "iteioctl.h"
#include "machine/cpu.h"

#include "../amiga/cc_types.h"
#include "../amiga/cc_chipmem.h"

#include "grf/grf_types.h"
#include "grf/grf_view.h"
#include "grf/grf_bitmap.h"
#include "viewioctl.h"
#include "viewvar.h"

extern unsigned char kernel_font_width, kernel_font_height, kernel_font_baseline; 
extern short         kernel_font_boldsmear;
extern unsigned char kernel_font_lo, kernel_font_hi;
extern unsigned char kernel_font[], kernel_cursor[];

/*
 * This is what ip->priv points to;
 * it contains local variables for custom-chip ites.
 */
typedef struct ite_priv {
  view_t *view;			/* the view for this ite. */
  u_char **row_ptr;		/* array of pointers into the bitmap  */
  u_long row_bytes;
  u_long cursor_opt;
} ipriv_t;

extern struct itesw itesw[];

/* (M<=8)-by-N routines */
static void view_le32n_cursor(struct ite_softc *ip, int flag);
static void view_le8n_putc(struct ite_softc *ip, int c, int dy, int dx, int mode);
static void view_le8n_clear(struct ite_softc *ip, int sy, int sx, int h, int w);
static void view_le8n_scroll(struct ite_softc *ip, int sy, int sx, int count, int dir);
void scroll_bitmap (bmap_t *bm, u_word x, u_word y, u_word width, u_word height, word dx, word dy, u_byte mask);

/* globals */
int ite_default_width = 640;
#if defined (GRF_NTSC)
int ite_default_height = 400;
int ite_default_depth = 2;
#elif defined (GRF_PAL)
int ite_default_height = 512;
int ite_default_depth = 2;
#else
int ite_default_height = 400;
int ite_default_depth = 2;
#endif

/* audio bell stuff */
static char *bell_data;
static struct ite_bell_values bell_vals = { 64, 200, 10 };

cc_unblank ()
{
}

init_bell ()
{
    short i;
    static char sample[20] = {
	0,39,75,103,121,127,121,103,75,39,0,
	-39,-75,-103,-121,-127,-121,-103,-75,-39 };

    if (!bell_data) {
	bell_data = alloc_chipmem(20);
	if (!bell_data)
	   return (-1); 

	for (i=0; i<20; i++) 
	    bell_data[i] = sample[i];
    }
    return (0);
}

cc_bell ()
{
    if (bell_data) {
      play_sample (10, PREP_DMA_MEM(bell_data),
                   bell_vals.period, bell_vals.volume, 0x3, bell_vals.time);
    }
}

extern struct  ite_softc ite_softc[];
#define IPUNIT(ip) (((u_long)ip-(u_long)ite_softc)/sizeof(struct ite_softc))

int 
ite_new_size (struct ite_softc *ip, struct ite_window_size *vs)
{
    extern struct view_softc views[];
    ipriv_t *cci = ip->priv;    
    u_long fbp, i;
    int error;

    error = viewioctl (IPUNIT (ip), VIEW_SETSIZE, (struct view_size *)vs, 0, -1);
    cci->view = views[IPUNIT(ip)].view; 

    ip->rows = cci->view->display.height / ip->ftheight;
    ip->cols = cci->view->display.width  / ip->ftwidth;

    if (cci->row_ptr) {
        free_chipmem (cci->row_ptr);
    }

    cci->row_ptr = alloc_chipmem (sizeof (u_byte *)*ip->rows);
    if (!cci->row_ptr) {
	panic ("no memory for console device.");
    }

    cci->row_bytes = (cci->view->bitmap->bytes_per_row + cci->view->bitmap->row_mod); 
    cci->row_bytes *= ip->ftheight;
    fbp = VDISPLAY_LINE (cci->view, 0, 0);
    for (i = 0; i < ip->rows; i++) {
	cci->row_ptr[i] = (u_char*)fbp;
	fbp += cci->row_bytes; 
    }
    return (error);
}


ite_view_init(ip)
	register struct ite_softc *ip;
{
    struct itesw *sp = itesw;
    ipriv_t *cci = ip->priv;
    struct ite_window_size vs;

    init_bell ();

    if (!cci) {
	ip->font     = kernel_font;
	ip->font_lo  = kernel_font_lo;
	ip->font_hi  = kernel_font_hi;
	ip->ftwidth  = kernel_font_width;
	ip->ftheight = kernel_font_height;
	ip->ftbaseline = kernel_font_baseline;
	ip->ftboldsmear = kernel_font_boldsmear;

	/* Find the correct set of rendering routines for this font.  */
	if (ip->ftwidth <= 8) {
		sp->ite_cursor = (void*)view_le32n_cursor;
		sp->ite_putc = (void*)view_le8n_putc;
		sp->ite_clear = (void*)view_le8n_clear;
		sp->ite_scroll = (void*)view_le8n_scroll;
	} else { 
		panic("kernel font size not supported");
	}

	cci = alloc_chipmem (sizeof (*cci));
	if (!cci) {
		panic ("no memory for console device.");
	}

	ip->priv = cci;

 	cci->cursor_opt = 0;
	cci->view = NULL;
	cci->row_ptr = NULL;

#if 0
	/* currently the view is permanently open due to grf driver */
	if (viewopen (IPUNIT(ip), 0)) {
	    panic ("cannot get ahold of our view");
	}
#endif

	vs.width = ite_default_width;
	vs.height = ite_default_height;
	vs.depth = ite_default_depth;

	ite_new_size (ip, &vs);
	viewioctl (IPUNIT(ip), VIEW_DISPLAY, NULL, 0, -1);
    }
}

int
ite_grf_ioctl (ip, cmd, addr, flag)
  struct ite_softc *ip;
  caddr_t addr; 
{
    int error = 0;
    struct ite_window_size *is;
    struct ite_bell_values *ib;
    ipriv_t *cci = ip->priv;

    switch (cmd) {
      case ITE_GET_WINDOW_SIZE:
	is = (struct ite_window_size *)addr;

	is->width = cci->view->display.width;
 	is->height = cci->view->display.height;
 	is->depth = cci->view->bitmap->depth;
        break;

      case ITE_SET_WINDOW_SIZE:
        is = (struct ite_window_size *)addr;

        if (ite_new_size (ip, is)) {
            error = ENOMEM;
        } else {
	    ite_reset (ip);
        }
        break;

      case ITE_DISPLAY_WINDOW:
	viewioctl (IPUNIT(ip), VIEW_DISPLAY, NULL, 0, -1);
	break;

      case ITE_REMOVE_WINDOW:
        viewioctl (IPUNIT(ip), VIEW_REMOVE, NULL, 0, -1);
        break;

      case ITE_GET_BELL_VALUES:
      	ib = (struct ite_bell_values *)addr;
	bcopy (&bell_vals, ib, sizeof (struct ite_bell_values));	
	break;

      case ITE_SET_BELL_VALUES:
      	ib = (struct ite_bell_values *)addr;
	bcopy (ib, &bell_vals, sizeof (struct ite_bell_values));	
	break;

      default:
        error = -1;
        break;
    }
    return (error);
}

ite_view_deinit(ip)
	struct ite_softc *ip;
{
  ip->flags &= ~ITE_INITED;
  /* FIX: free our stuff. */
  if (ip->priv) {
    ipriv_t *cci = ip->priv;

#if 0
    /* view stays open permanently */
    if (cci->view) {
        viewclose (IPUNIT(ip));
    }
#endif

    if (cci->row_ptr) {
   	free_chipmem (cci->row_ptr);
    }
    free_chipmem (cci);
    ip->priv = NULL;
  }
}

/*
 * Swap two text rows in the display by modifying the copper list.
 */
static __inline int
swap_rows(struct ite_softc *ip, int row1, int row2)
{
	/* If the drawn cursor was on either row, swap it too.  */
	if (ip->cursory == row1)
		ip->cursory = row2;
	else if (ip->cursory == row2)
		ip->cursory = row1;
}

/*** (M<8)-by-N routines ***/
#define CO_DRAWN 0x1
#define CO_OPT 0x2

/* NOTE: This routine assumes a cursor overlay plane,
   but it does allow cursors up to 32 pixels wide.  */
static void
view_le32n_cursor(struct ite_softc *ip, int flag)
{
	ipriv_t  *cci = (ipriv_t *) ip->priv;
	view_t *v = cci->view;
	bmap_t *bm = v->bitmap;
	int dr_plane = (bm->depth > 1 ? bm->depth-1 : 0);
	int cend, ofs, h, cstart;
	unsigned char opclr, opset;
	u_char *pl;

	if (flag == START_CURSOROPT || flag == END_CURSOROPT) 
	    return;

	cstart = (cci->cursor_opt&CO_OPT ? ip->ftheight-1 : 0);
	cend = (cci->cursor_opt&CO_OPT ? 0 : ip->ftheight-1); 
	pl = VDISPLAY_LINE (v, dr_plane, (ip->cursory*ip->ftheight+cstart));
	ofs = v->display.x + (ip->cursorx * ip->ftwidth);

	/* erase cursor */
	if (flag != DRAW_CURSOR /* && (cci->cursor_opt & CO_DRAWN)*/) {
	    /* erase the cursor */
	    int h;
	    if (dr_plane) {
	   	for (h = cend; h >= 0; h--) {
		    asm("bfclr %0@{%1:%2}"
		      : : "a" (pl), "d" (ofs), "d" (ip->ftwidth));
		    pl += bm->row_mod + bm->bytes_per_row;
	        }
	    } else {
	    	for (h = cend; h >= 0; h--) {
		    asm("bfchg %0@{%1:%2}"
		      : : "a" (pl), "d" (ofs), "d" (ip->ftwidth));
		    pl += bm->row_mod + bm->bytes_per_row;
	    	}
	    }
	    cci->cursor_opt &= ~CO_DRAWN;
    	}

	/* set new height */

	if (flag == START_CURSOROPT) {
	    cci->cursor_opt |= CO_OPT;
	} else if (flag == END_CURSOROPT) {
	    cci->cursor_opt &= ~CO_OPT;
	}


	if ((flag == DRAW_CURSOR ||
	                                        flag == MOVE_CURSOR ||
	                                        flag == START_CURSOROPT ||
	                                        flag == END_CURSOROPT)) {
	ip->cursorx = MIN(ip->curx, ip->cols-1);
	ip->cursory = ip->cury;
	cstart = (cci->cursor_opt & CO_OPT ? ip->ftheight-1 : 0);
	cend = (cci->cursor_opt & CO_OPT ? 0 : ip->ftheight-1); 
	pl = VDISPLAY_LINE (v, dr_plane, (ip->cursory*ip->ftheight+cstart));
	ofs = v->display.x + (ip->cursorx * ip->ftwidth);
	/* if (!(cci->cursor_opt & CO_DRAWN)) { */
	    /* draw the cursor */
	    if (dr_plane) {
		for (h = cend; h >= 0; h--) {
		    asm("bfset %0@{%1:%2}"
		      : : "a" (pl), "d" (ofs), "d" (ip->ftwidth));
		    pl += bm->row_mod + bm->bytes_per_row;
	    	}
	    } else {
		for (h = cend; h >= 0; h--) {
		    asm("bfchg %0@{%1:%2}"
		      : : "a" (pl), "d" (ofs), "d" (ip->ftwidth));
		    pl += bm->row_mod + bm->bytes_per_row;
	    	}
	    }
	    cci->cursor_opt |= CO_DRAWN;
	/* } */
	}
}


static inline
int expbits (int data)
{
    int i, nd = 0;
    if (data & 1) {
    	nd |= 0x02;
    }
    for (i=1; i < 32; i++) {
	if (data & (1 << i)) {
	   nd |= 0x5 << (i-1);
	}
    }
    nd &= ~data;
    return (nd);
}

/* FIX: shouldn't this advance the cursor even if the character to
        be output is not available in the font? -ch */

static void
view_le8n_putc(struct ite_softc *ip, int c, int dy, int dx, int mode)
{
  if (c >= ip->font_lo && c <= ip->font_hi)
    {
      ipriv_t *cci = (ipriv_t *) ip->priv;
      bmap_t *bm = cci->view->bitmap;
      register fh = ip->ftheight, fw = ip->ftwidth;
      register rofs = bm->row_mod + bm->bytes_per_row;
      u_char *pl = cci->row_ptr[dy];
      int ofs = dx * ip->ftwidth, width = bm->bytes_per_row << 3;
      int baseline = ip->ftbaseline+1; /* FIX: XXX this is for my font :^) */
      u_char *fontp = ip->font + (c - ip->font_lo) * ip->ftheight;
      int h;

      if (mode & ATTR_INV)
	{
  	  if (mode & ATTR_BOLD)
	    {
	      if (fw+ofs+2 < width) 
	        fw++;

              for (h=0; h < fh; h++)
	        {
	          int ch;
	          ch = fontp[h];
	          ch = ch|(ch<<1); 

	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (~(ch)));
	          pl += rofs;
	        }
	      if (mode & ATTR_UL) 
	        {
	          int ch = expbits (fontp[baseline]|(fontp[baseline]<<1));

	          pl = cci->row_ptr[dy];
	          pl += (baseline)*(rofs);

 	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (ch));
	        }

	    } 
	  else
	    {
              for (h=0; h < fh; h++)
	        {
	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (~fontp[h]));
	          pl += bm->row_mod + bm->bytes_per_row; 
	        }
	      if (mode & ATTR_UL) 
	        {
	          int ch = expbits (fontp[baseline]);

	          pl = cci->row_ptr[dy];
	          pl += (baseline)*(rofs);

 	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (ch));
	        }
	    }
	}	
      else /* NORMAL */
        {
       	  if (mode & ATTR_BOLD)
	    {
	      if (fw+ofs+2 < width) 
	        fw++;

              for (h=0; h < fh; h++)
	        {
	          int ch = fontp[h];
	          ch = ch|(ch<<1); 

	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (ch));
	          pl += rofs;
	        }
	      if (mode & ATTR_UL) 
	        {
	          int ch = ~expbits (fontp[baseline]|(fontp[baseline]<<1));

	          pl = cci->row_ptr[dy];
	          pl += (baseline)*(rofs);

 	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (ch));
	        }

	    } 
	  else
	    {
              for (h=0; h < fh; h++)
	        {
	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (fontp[h]));
	          pl += bm->row_mod + bm->bytes_per_row; 
	        }
	      if (mode & ATTR_UL) 
	        {
	          int ch = ~expbits (fontp[baseline]);

	          pl = cci->row_ptr[dy];
	          pl += (baseline)*(rofs);

 	          asm("bfins %3,%0@{%1:%2}"
	              : : "a" (pl), "d" (ofs), "d" (fw), "d" (ch));
	        }
	    }
	}
    }
}

static void
view_le8n_clear(struct ite_softc *ip, int sy, int sx, int h, int w)
{
  ipriv_t *cci = (ipriv_t *) ip->priv;
  view_t *v = cci->view;
  bmap_t *bm = cci->view->bitmap;

  if ((sx == 0) && (w == ip->cols))
    {
      /* common case: clearing whole lines */
      while (h--)
	{
	  int i;
	  u_char *ptr = cci->row_ptr[sy]; 
	  for (i=0; i < ip->ftheight; i++) {
            bzero(ptr, bm->bytes_per_row);
            ptr += bm->bytes_per_row + bm->row_mod;			/* don't get any smart
                                                   ideas, becuase this is for
                                                   interleaved bitmaps */
          }
	  sy++;
	}
    }
  else
    {
      /* clearing only part of a line */
      /* XXX could be optimized MUCH better, but is it worth the trouble? */
      while (h--)
	{
	  u_char *pl = cci->row_ptr[sy];
          int ofs = sx * ip->ftwidth;
	  int i, j;
	  for (i = w-1; i >= 0; i--)
	    {
	      u_char *ppl = pl;
              for (j = ip->ftheight-1; j >= 0; j--)
	        {
	          asm("bfclr %0@{%1:%2}"
	              : : "a" (ppl), "d" (ofs), "d" (ip->ftwidth));
	          ppl += bm->row_mod + bm->bytes_per_row; 
	        }
	      ofs += ip->ftwidth;
	    }
	  sy++;
	}
    }
}

/* Note: sx is only relevant for SCROLL_LEFT or SCROLL_RIGHT.  */
static void
view_le8n_scroll(ip, sy, sx, count, dir)
        register struct ite_softc *ip;
        register int sy;
        int dir, sx, count;
{
  bmap_t *bm = ((ipriv_t *)ip->priv)->view->bitmap;
  u_char *pl = ((ipriv_t *)ip->priv)->row_ptr[sy];

  if (dir == SCROLL_UP) 
    {
      int dy = sy - count;
      int height = ip->bottom_margin - sy + 1;
      int i;

      /*FIX: add scroll bitmap call */
      view_le32n_cursor(ip, ERASE_CURSOR);
	scroll_bitmap (bm, 0, dy*ip->ftheight,
		       bm->bytes_per_row >> 3, (ip->bottom_margin-dy+1)*ip->ftheight,
		       0, -(count*ip->ftheight), 0x1);
/*	if (ip->cursory <= bot || ip->cursory >= dy) {
	    ip->cursory -= count;
	} */
    }
  else if (dir == SCROLL_DOWN) 
    {
      int dy = sy + count;
      int height = ip->bottom_margin - dy + 1;
      int i;

      /* FIX: add scroll bitmap call */
      view_le32n_cursor(ip, ERASE_CURSOR);
	scroll_bitmap (bm, 0, sy*ip->ftheight,
		       bm->bytes_per_row >> 3, (ip->bottom_margin-sy+1)*ip->ftheight,
		       0, count*ip->ftheight, 0x1);
/*	if (ip->cursory <= bot || ip->cursory >= sy) {
	    ip->cursory += count;
	} */
    }
  else if (dir == SCROLL_RIGHT) 
    {
      int sofs = (ip->cols - count) * ip->ftwidth;
      int dofs = (ip->cols) * ip->ftwidth;
      int i, j;

      view_le32n_cursor(ip, ERASE_CURSOR);
      for (j = ip->ftheight-1; j >= 0; j--)
	{
	  int sofs2 = sofs, dofs2 = dofs;
	  for (i = (ip->cols - (sx + count))-1; i >= 0; i--)
	    {
	      int t;
	      sofs2 -= ip->ftwidth;
	      dofs2 -= ip->ftwidth;
	      asm("bfextu %1@{%2:%3},%0"
	          : "=d" (t)
		  : "a" (pl), "d" (sofs2), "d" (ip->ftwidth));
	      asm("bfins %3,%0@{%1:%2}"
	          : : "a" (pl), "d" (dofs2), "d" (ip->ftwidth), "d" (t));
	    }
	  pl += bm->row_mod + bm->bytes_per_row; 
	}
    }
  else /* SCROLL_LEFT */
    {
      int sofs = (sx) * ip->ftwidth;
      int dofs = (sx - count) * ip->ftwidth;
      int i, j;

      view_le32n_cursor(ip, ERASE_CURSOR);
      for (j = ip->ftheight-1; j >= 0; j--)
	{
	  int sofs2 = sofs, dofs2 = dofs;
	  for (i = (ip->cols - sx)-1; i >= 0; i--)
	    {
	      int t;
	      asm("bfextu %1@{%2:%3},%0"
	          : "=d" (t)
		  : "a" (pl), "d" (sofs2), "d" (ip->ftwidth));
	      asm("bfins %3,%0@{%1:%2}"
	          : : "a" (pl), "d" (dofs2), "d" (ip->ftwidth), "d" (t));
	      sofs2 += ip->ftwidth;
	      dofs2 += ip->ftwidth;
	    }
	  pl += bm->row_mod + bm->bytes_per_row; 
	}
    }		
}

void inline 
scroll_bitmap_l (bmap_t *bm, u_word x, u_word y, u_word width, u_word height, word dx, word dy, u_byte mask)
{
    u_word depth = bm->depth; 
    u_word lwpr = bm->bytes_per_row >> 2;
    if (dx) {
    	/* FIX: */ panic ("delta x not supported in scroll bitmap yet.");
    } 
    if (bm->flags & BMF_INTERLEAVED) {
	height *= depth;
	depth = 1;
    }
    if (dy == 0) {
        return;
    }
    if (dy > 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
	    	u_long *pl = (u_long *)bm->plane[i];
		u_long *src_y = pl + (lwpr*y);
		u_long *dest_y = pl + (lwpr*(y+dy));
		u_long count = lwpr*(height-dy);
		u_long *clr_y = src_y;
		u_long clr_count = dest_y - src_y;
	
		src_y += count - 1;
		dest_y += count - 1;

		while (count--) {
			*dest_y-- = *src_y--;
		}
		while (clr_count--) {
		    	*clr_y++ = 0;
		}
    	    }
	}
    } else if (dy < 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
    		u_long *pl = (u_long *)bm->plane[i];
    		u_long *src_y = pl + (lwpr*(y-dy));
    		u_long *dest_y = pl + (lwpr*y); 
		long count = lwpr*(height + dy);
		u_long *clr_y = dest_y + count;
		u_long clr_count = src_y - dest_y;

		while (count--) {
		    *dest_y++ = *src_y++;
		}
		while (clr_count--) {
		    *clr_y++ = 0;
		}
	    }
	}
    }
}

void inline
scroll_bitmap_w (bmap_t *bm, u_word x, u_word y, u_word width, u_word height, word dx, word dy, u_byte mask)
{
    u_word depth = bm->depth; 
    u_word wpr = bm->bytes_per_row >> 1;

    if (dx) {
    	/* FIX: */ panic ("delta x not supported in scroll bitmap yet.");
    } 
    if (bm->flags & BMF_INTERLEAVED) {
	height *= depth;
	depth = 1;
    }
    if (dy == 0) {
        return;
    }
    if (dy > 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
	    	u_word *pl = (u_word *)bm->plane[i];
		u_word *src_y = pl + (wpr*y);
		u_word *dest_y = pl + (wpr*(y+dy));
		u_long count = wpr*(height-dy);
		u_word *clr_y = src_y;
		u_long clr_count = dest_y - src_y;
	
		src_y += count - 1;
		dest_y += count - 1;

		while (count--) {
			*dest_y-- = *src_y--;
		}
		while (clr_count--) {
		    	*clr_y++ = 0;
		}
    	    }
	}
    } else if (dy < 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
    		u_word *pl = (u_word *)bm->plane[i];
    		u_word *src_y = pl + (wpr*(y-dy));
    		u_word *dest_y = pl + (wpr*y); 
		u_long count = wpr*(height + dy);
		u_word *clr_y = dest_y + count;
		u_long clr_count = src_y - dest_y;

		while (count--) {
		    *dest_y++ = *src_y++;
		}
		while (clr_count--) {
		    *clr_y++ = 0;
		}
	    }
	}
    }
}

void inline
scroll_bitmap_b (bmap_t *bm, u_word x, u_word y, u_word width, u_word height, word dx, word dy, u_byte mask)
{
    u_word depth = bm->depth; 
    u_word bpr = bm->bytes_per_row;

    if (dx) {
    	/* FIX: */ panic ("delta x not supported in scroll bitmap yet.");
    } 
    if (bm->flags & BMF_INTERLEAVED) {
	height *= depth;
	depth = 1;
    }
    if (dy == 0) {
        return;
    }
    if (dy > 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
	    	u_byte *pl = (u_byte *)bm->plane[i];
		u_byte *src_y = pl + (bpr*y);
		u_byte *dest_y = pl + (bpr*(y+dy));
		u_long count = bpr*(height-dy);
		u_byte *clr_y = src_y;
		u_long clr_count = dest_y - src_y;
	
		src_y += count - 1;
		dest_y += count - 1;

		while (count--) {
			*dest_y-- = *src_y--;
		}
		while (clr_count--) {
		    	*clr_y++ = 0;
		}
    	    }
	}
    } else if (dy < 0) {
    	int i;
    	for (i=0; i < depth && mask; i++, mask >>= 1) {
    	    if (0x1 & mask) {
    		u_byte *pl = (u_byte *)bm->plane[i];
    		u_byte *src_y = pl + (bpr*(y-dy));
    		u_byte *dest_y = pl + (bpr*y); 
		u_long count = bpr*(height + dy);
		u_byte *clr_y = dest_y + count;
		u_long clr_count = src_y - dest_y;

		while (count--) {
		    *dest_y++ = *src_y++;
		}
		while (clr_count--) {
		    *clr_y++ = 0;
		}
	    }
	}
    }
}

/* right now the only thing you can do is scroll up or down */
/* all allocations from alloc_bitmap return PAGE aligned bitmap data. */
/* not all bitmaps are aligned to long words. (i.e. bpr is expressable in lw's */

/* the deltas represent what to add to there respective points */ 
/* so negative is left or up for x and y */

/* FIX: XXX currently does not support masks and interleaved bitmaps */

void
scroll_bitmap (bmap_t *bm, u_word x, u_word y, u_word width, u_word height, word dx, word dy, u_byte mask)
{
    switch (bm->bytes_per_row & 0x3) {
      case 0:
	/* long word aligned */
	return (scroll_bitmap_l (bm,x,y,width,height,dx,dy,mask));
      case 2:
      	/* word aligned */
	return (scroll_bitmap_w (bm,x,y,width,height,dx,dy,mask));
      default:
        /* byte aligned */
	return (scroll_bitmap_w (bm,x,y,width,height,dx,dy,mask));
    }
}

#endif
